/**************************************************************************

   File:                PCA.cpp

   Author(s):           Robert Sim

   Created:             11 Nov 1997

   Last Revision:       $Date: 1998/07/21 14:20:29 $

   Description: A Class for performing Principal Components Decomposition

   $Revision: 2.0 $

   $Log: PCA.cpp,v $
   Revision 2.0  1998/07/21 14:20:29  simra
   Final Thesis submitted version.

   Revision 1.10  1998/04/14 04:29:48  robbo
   Debugging changes.

   Revision 1.9  1998/04/14 03:38:06  robbo
   Altered PCA.  These changes should either be undone, or continued.

   Revision 1.8  1998/04/07 23:38:25  simra
   Changes to compile on sgi w/ CC. It works- huge speedup!

   Revision 1.7  1998/04/04 06:11:07  robbo
   Fixed filtering. Sm and Cheeseman is a bit too enthusiastic right now-
   the covariances are too small.  Avg error rate on scene.gml (0.01,0.2)!

   Revision 1.6  1998/04/03 23:26:01  simra
   Debugging outlier removal. Still.

   Revision 1.5  1998/03/26 20:38:42  robbo
   Abandoned the .crt file format.  Now when landmarks are tracked the
   certainty is computed on the spot.  This will save some headaches.

   TODO: Get normalization functional for localization.

   Revision 1.4  1998/03/18 17:33:26  simra
   1. alpha failed...
   2. matrix and vector reads still need work.

   Revision 1.3  1998/03/18 15:10:34  robbo
   1. Fixed matrix&vector reads.
   2. Fixed Statistic reads.
   3. Fixed the 75% bug- no need for iteration now.
   4. Added -w flag for localize matching.

   Revision 1.2  1998/02/24 13:35:06  robbo
   Changed eigentrack extensively to update the pca on the fly- best
   distinguishing the tracked sets. This required some major overhauling.
   Also changed some input file conventions. landmark outputs a clist,
   eigentrack inputs the same clist. eigenland is no longer useful for
   building the database.  The separate clist prevents the clist in the
   database from having to be recomputed every time (saving a lot of
   trouble.)

   Revision 1.1.1.1  1998/02/24 02:12:11  robbo
   First fie import.


	
   Copyright (c) 1997 by Robert Sim, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#define PCA_CPP
#include "PCA.h"
PCA::PCA() {
  test_flag=false;
}

PCA::~PCA() {
}

inline void
PCA::ROTATE(matrix& a, const int i, const int j, const int k, const int l) {
  g=a(i,j);
  h=a(k,l);
  a(i,j)=g-s*(h+g*tau);
  a(k,l)=h+s*(g-h*tau);
}

double
PCA::CONDITION_THRESH(1e-3);

PrincipalComponents
PCA::GetSVDPrincipalComponents(matrix &P, bool norm) {
  vector average=vector(P.dim2());
  if (norm) {
    verbprintf(3,"\tNormalizing\n");
    int i;
    for (i=0;i<P.dim1();i++) {
      verbprintf(3, "\tRow: %d\r",i);
      P[i] = P[i].norm();
    }
  }
  verbprintf(3, "\n");
  verbprintf(3,"\tAveraging\n");
  int i;
  for (i=0; i<P.dim1(); i++) {
    average+=P[i];
    verbprintf(3, "\tAvg Row: %d\r",i);
  }
  average=average/(double)P.dim1();
  verbprintf(3, "\n\tTranslating\n");
  for (i=0;i<P.dim1(); i++) {
    verbprintf(3, "\tTr Row: %d\r", i);
    P[i]-=average;
  }

  // Augment the matrix with zero columns if the system is overdetermined.
  //int origdim1=P.dim1();
  int origdim2=P.dim2();
   if (P.dim2()<P.dim1()) {
     matrix Pnew(P.dim1(),P.dim1());
     int r,c;
     for (r=0; r<P.dim1(); r++) {
       for (c=0; c<P.dim2(); c++) {
	 Pnew(r,c)=P(r,c);
       }
     }
     P=Pnew;
   }

  matrix U;
  vector w;
  matrix V;
  P.trans().svdcmp(U,w,V);
  // remove zero w's.
  if (w[0]<P.dim1()*MINDOUBLE) {
    cerr << "Trouble computing SVD"<<endl;
    return PrincipalComponents();
  }
  int numvects=0;
  double thresh=CONDITION_THRESH; // conditioning number..
  int k,l;
  int biggest=0;

  for (k=1; k<w.dim(); k++) {
    if (w[biggest]<w[k]) biggest=k;
  }

  // This is where the value of K is determined
  verbprintf(3,"\nEigenValues:\n");
  for (k=0; k<w.dim(); k++) {
    verbprintf(3,"%d:%lf\n",k,w[k]);
    if (w[k]/w[biggest]<thresh) w[k]=0;
    else numvects++;
  }
  // remove zeroed cols of U and transpose U into Uresult.
  matrix Uresult(numvects, origdim2); // Uresult.dim1() is K ie numvects
  vector newW(numvects);
  int thisrow=0;
  for (k=0; k<w.dim(); k++) {
    if (w[k]!=0) {
      newW[thisrow]=w[k];
      for (l=0; l<origdim2; l++) {
	Uresult(thisrow,l)=U(l,k);
      }
      thisrow++;
    }
  }
  
  /*  
      cout << "Vector of EigenValues\n";
      cout << newW << "\n";
  */

  verbprintf(3, "K: %d N: %d\n", Uresult.dim1(), Uresult.dim2());
  /*  
      cout << "Uresult= " << Uresult << "\n";
      cout << "average= " << average << "\n";
      cout << "newW= " << newW << "\n";
      cout << "norm= " << norm << "\n";
  */
  return PrincipalComponents(Uresult,average,newW,norm);
}


PrincipalComponents
PCA::GetPrincipalComponents(matrix &P,bool norm) {
  verbprintf(2,"Performing PCA..\n");
  vector average(P.dim2());
  if (norm) {
    verbprintf(2,"\tNormalizing\n");
    int i;
    for (i=0;i<P.dim1();i++) {
      verbprintf(2, "\tRow: %d\r",i);
      P.row(i) = P.row(i).norm();
    }
  }
  verbprintf(2, "\n");
  verbprintf(2,"\tAveraging and translating\n");
  int i;
  for (i=0; i<P.dim1(); i++) {
    average+=P.row(i);
    verbprintf(2, "\tAvg Row: %d\r",i);
  }
  average=average/(double)P.dim1();
  verbprintf(2, "\n");
  for (i=0;i<P.dim1(); i++) {
    verbprintf(2, "\tTr Row: %d\r", i);
    P.row(i)-=average;

  }
  verbprintf(2, "\n");
  verbprintf(2, "\tFinding Eigenvectors of Qimp\n");
  verbprintf(2, "\tBuilding Q\n");
  matrix Q=P*P.trans();
  
  vector b(Q.dim1());
  vector Qevals(Q.dim1());
  matrix Qevects(Q.dim1(),Q.dim1());
  
  for (i=0; i<Q.dim1(); i++) {
    verbprintf(2, "\tQ Row %d\r", i);
    Qevects(i,i)=1.0;
    b[i]=Qevals[i]=Q(i,i);
  }
  verbprintf(2, "\n");
  /*  if (Q!=Q.trans()) {
      printf(1,"Doh!\n");
      }*/

  int nrot;
  double sm=0.0;
  for (nrot=1; nrot<=50; nrot++) {
    verbprintf(2, "\tIteration %d\r", nrot);
    sm=0.0;
    int p,q;
    for (p=0; p<Q.dim1()-1;p++) { // (n-1???!!!??!)
      for (q=p+1; q<Q.dim1();q++) {
	sm +=fabs(Q(p,q));
      }
    }
    
    if (sm==0.0) {
      verbprintf(2, "\nDone\n");
      break;
    } else {
      vector z(Q.dim1());
      double thresh;
      
      if (nrot<4) thresh=0.2*sm/(Q.dim1()*Q.dim1());
      else thresh=0.0;
      
      for (p=0; p<Q.dim1()-1; p++) {
	for (q=p+1;q<Q.dim1();q++) {
	  g=100.0*fabs(Q(p,q));
	  if (nrot>4 && (fabs(Qevals[p])+g==fabs(Qevals[p]))
	      && (fabs(Qevals[q]+g==fabs(Qevals[q]))))
	    Q(p,q)=0.0;
	  else if (fabs(Q(p,q))>thresh) {
	    h=Qevals[q]-Qevals[p];
	    double t;
	    if (fabs(h)+g==fabs(h))
	      t=Q(p,q)/h;
	    else {
	      double theta=0.5*h/Q(p,q);
	      t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
	      if (theta<0.0) t=-t;
	    }
	    double c=1.0/sqrt(1+t*t);
	    s=t*c;
	    tau=s/(1.0+c);
	    h=t*Q(p,q);
	    z[p]-=h;
	    z[q]+=h;
	    Qevals[p]-=h;
	    Qevals[q]+=h;
	    Q(p,q)=0.0;
	    int j;
	    for (j=0;j<=p-1;j++) {
	      ROTATE(Q,j,p,j,q);
	    }
	    for (j=p+1;j<=q-1;j++) {
	      ROTATE(Q,p,j,j,q);
	    }
	    for (j=q+1;j<Q.dim1();j++) {
	      ROTATE(Q,p,j,q,j);
	    }
	    for (j=0;j<Q.dim1();j++) {
	      ROTATE(Qevects, j,p,j,q);
	    } 
	  }
	}
      }
      b+=z;
      Qevals=b;
    }
  }
  verbprintf(3, "Nrot: %d\n", nrot);
  if (nrot==51) {
    verbprintf(1, "\nToo many rotations in jacobi routine %f\n",sm);
    return PrincipalComponents();
  }

  //  printf(3, "\n");
  verbprintf(2, "\tSorting Eigenvectors\n");
  //  vector p=Qevals;
  double p;
  int j,k;
  for (i=0;i<Q.dim1()-1;i++) {
    p=Qevals[k=i];
    for (j=i+1;j<Q.dim1();j++)
      if (Qevals[j] >= p) p=Qevals[k=j];
    if (k != i) {
      Qevals[k]=Qevals[i];
      Qevals[i]=p;
      for (j=0;j<Q.dim1();j++) {
	p=Qevects(j,i);
	Qevects(j,i)=Qevects(j,k);
	Qevects(j,k)=p;
      }
    }
  }


  
  verbprintf(2, "\tTransforming eigenvectors of Qimp to eigenvectors of Q\n");
  matrix Evects=(P.trans()*Qevects).trans(); //MxN

  verbprintf(2, "\tnormalizing\n");
  for (i=0; i<Evects.dim1(); i++) {
    Evects.row(i)=Evects.row(i).norm();
  }

  if (test_flag) {
    matrix I=Evects*Evects.trans();
    verbprintf(1,"dots = evc*evc^t\n");
    I.print();
  }
  return PrincipalComponents(Evects,average,Qevals,norm);
}




