/************************************************************************

   File:		fft.cpp

   Author(s):		Deeptiman Jugessur

   Created:		14 Jul 1999

   Last Revision:	$Date$

   Description:	

   $Revision$

   $Log$

	
   Copyright (c) 1999 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 ***********************************************************************/

#include "fft.h"

#ifdef SWAP
#error "Can't define swap"
#endif

#define SWAP(a, b) {double tempr=(a); (a)=(b); (b)=tempr;}



fft::fft()
{
 re = NULL; 
 im = NULL; 
 shiftRe = NULL;
 shiftIm = NULL;
 amplitude = NULL; 
 phase = NULL;
}

/*
  Given a matrix whose dimesions have to be a power of 2, this constructor
  does a 2 dimensional fft on the data, creates 4 new matrices of the same 
  size (actually would be two complex matrices).  The first 'complex' 
  matrix is that of the frequency values such that the zero frequency 
  components lie in the top left corner.  The second 'complex' matrix is the
  more useful one where the 0 frequency components are in the center.
*/
fft::fft(const matrix& in)
{
  int numRowBits=0;
  int numColBits=0;
  int numRows=in.dim1();
  int numCols=in.dim2();
  int halfNumRows = numRows/2;
  int halfNumCols = numCols/2;

  
  // check that they are both powers of 2
  int copy=numRows;
  int andVal=0;
  while (copy != 0) {
    andVal = numRows & copy;
    if (andVal)
      numRowBits++;
    copy >>= 1;
  }

  copy=numCols;
  andVal=0;
  while (copy != 0) {
    andVal = numCols & copy;
    if (andVal)
      numColBits++;
    copy >>= 1;
  }

  //   cout << numRowBits << "," << numColBits << "\n";

  if (numRowBits != 1 || numColBits != 1)
    nrerror("Cannot perform FFT on dimensions which are not powers of 2\n");

  float ***data, **speq;
  // use nr space allocater
  data = f3tensor(1,1,1,numRows,1,numCols);
  speq = nrmatrix(1,1,1,2*numRows);


  // load the matrix data into data, maybe I could be more efficient
  for (int i=1; i<=numRows; i++) {
    for (int j=1; j<=numCols; j++) {
      data[1][i][j] = in(i-1,j-1);
    }
  }
  
  // perform fft
  rlft3(data,speq,1,numRows,numCols,1);
  
  // allocate space for the private matrices which will store the data
  re        = new matrix(numRows, numCols);
  im        = new matrix(numRows, numCols);
  shiftRe   = new matrix(numRows, numCols);
  shiftIm   = new matrix(numRows, numCols);
  amplitude = NULL;
  phase     = NULL;
  
  for (int i=1; i<=numRows; i++) {
    for (int j=1; j<=halfNumCols; j++) {
      // cout << "data:(" << i << "," << j << ")--->Re = " << data[1][i][2*j-1] << ", im = " << data[1][i][2*j] << "\n";
      (*re)(i-1,j-1) =  data[1][i][2*j-1];
      (*im)(i-1,j-1) = -data[1][i][2*j];
    }
  }

  int count=1;
  for (int i=1; i<=numRows; i++) {
    // cout << "speq:(" << count << "," << numCols/2+1 << ")--->Re = " << speq[1][2*i-1] << ", im = " << speq[1][2*i] << "\n";
    (*re)(count-1,halfNumCols) =  speq[1][2*i-1];
    (*im)(count-1,halfNumCols) = -speq[1][2*i];
    count++;
  }

  // Fill in the first row entirely (different from others)
  for (int i=numCols-1; i>halfNumCols; i--) {
    (*re)(0,i) =  (*re)(0,numCols-i);
    (*im)(0,i) = -(*im)(0,numCols-i);
  }

  // Fill in the rest with symmetry properties
  for (int i=numRows-1; i>0; i--) {
    for (int j=numCols-1; j>halfNumCols; j--) {
      (*re)(i,j) =  (*re)(numRows-i,numCols-j);
      (*im)(i,j) = -(*im)(numRows-i,numCols-j);
    }
  }

  // Now want to swap the four quadrants such that 0 frequency values are in the middle.
  // since the mmatrices have to have dimesions of powers of 2, easily done.
  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      (*shiftRe)(i,j) = (*re)((i+halfNumRows) % numRows, (j+halfNumCols) % numCols);
      (*shiftIm)(i,j) = (*im)((i+halfNumRows) % numRows, (j+halfNumCols) % numCols);
    }
  }

  // free the nr stuff using nr code
  free_matrix(speq,1,1,1,2*numRows);
  free_f3tensor(data,1,1,1,numRows,1,numCols);

}


fft::~fft()
{
  if (re != NULL) 
    delete re;
  if (im != NULL)
    delete im;
  if (shiftRe != NULL)
    delete shiftRe;
  if (shiftIm != NULL)
    delete shiftIm;
  if (amplitude != NULL)
    delete amplitude;
  if (phase != NULL)
    delete phase;
}


void
fft::computeAmplitudeAndPhase()
{
  if (shiftRe == NULL || shiftIm == NULL) {
    fprintf(stderr,"Cannot compute amplitude and phase without fft\n");
    return;
  }

  if (amplitude != NULL)
    delete amplitude;

  if (phase != NULL)
    delete phase;

  // get the dimensions, shiftRe and shiftIm have the same dimensions
  int numRows = shiftRe->dim1();
  int numCols = shiftRe->dim2();

  amplitude = new matrix(numRows, numCols);
  phase     = new matrix(numRows, numCols);

  // fill up the amplitude and phase information
  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      (*amplitude)(i,j) = sqrt((*shiftRe)(i,j)*(*shiftRe)(i,j) + (*shiftIm)(i,j)*(*shiftIm)(i,j));
      (*phase)(i,j) = atan2((*shiftIm)(i,j),(*shiftRe)(i,j));
    }
  }
}


void 
fft::computeAmplitude()
{
  if (shiftRe == NULL || shiftIm == NULL) {
    fprintf(stderr,"Cannot compute amplitude without fft\n");
    return;
  }

  if (amplitude != NULL)
    delete amplitude;

  // get the dimensions, shiftRe and shiftIm have the same dimensions
  int numRows = shiftRe->dim1();
  int numCols = shiftRe->dim2();
  
  amplitude = new matrix(numRows, numCols);

  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      (*amplitude)(i,j) = sqrt((*shiftRe)(i,j)*(*shiftRe)(i,j) + (*shiftIm)(i,j)*(*shiftIm)(i,j));
    }
  }
}



void 
fft::computeLogAmplitude()
{
  if (shiftRe == NULL || shiftIm == NULL) {
    fprintf(stderr,"Cannot compute amplitude without fft\n");
    return;
  }

  if (amplitude != NULL)
    delete amplitude;

  // get the dimensions, shiftRe and shiftIm have the same dimensions
  int numRows = shiftRe->dim1();
  int numCols = shiftRe->dim2();
  
  amplitude = new matrix(numRows, numCols);

  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      (*amplitude)(i,j) = log(sqrt((*shiftRe)(i,j)*(*shiftRe)(i,j) + (*shiftIm)(i,j)*(*shiftIm)(i,j)));
      //      cout << "Before Log " << sqrt((*shiftRe)(i,j)*(*shiftRe)(i,j) + (*shiftIm)(i,j)*(*shiftIm)(i,j)) << " After Log " << (*amplitude)(i,j) << " at (" << i << "," << j << ")\n";
    }
  }
}

void 
fft::computePhase()
{
  if (shiftRe == NULL || shiftIm == NULL) {
    fprintf(stderr,"Cannot compute phase without fft\n");
    return;
  }

  if (phase != NULL)
    delete phase;

  // get the dimensions, shiftRe and shiftIm have the same dimensions
  int numRows = shiftRe->dim1();
  int numCols = shiftRe->dim2();

  phase = new matrix(numRows, numCols);

  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++) {
      (*phase)(i,j) = atan2((*shiftIm)(i,j),(*shiftRe)(i,j));
    }
  }
}


// given two phase images, this attempts to recover the phase difference
// between the two images by shifting one image over the other in theta 
// increments and performing a least squares fit each time.  The smallest 
// least squares value corresponds to the best angle.
double
fft::recoverPhase(matrix& M1, matrix& M2)
{
  assert(M1.dim1() == M2.dim1() && M1.dim2() == M2.dim2());

  double min=99999999;
  double ls=0;
  double result=0;
  float stepTheta=2*PI/M1.dim1();

  //  cout << "stepTheta = " << stepTheta << "\n";

  for (int i=0; i<M2.dim1(); i++) {
    ls = M1.leastSquares(M2,i,0);
    //    cout << i << ":ls = " << ls << "\n";
    if (ls <= min) {
      min = ls;
      result = i*stepTheta;
      //      cout << "Min found at " << i << "\n";
    }
  }

  return result;

}

// stright out of nr.  Comments would be useless :), its a magic black box function
void
fft::fourn(float data[], unsigned long nn[], int ndim, int isign)
{
  int idim;
  unsigned long i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
  unsigned long ibit,k1,k2,n,nprev,nrem,ntot;
  float tempi,tempr;
  double theta,wi,wpi,wpr,wr,wtemp;
  
  for (ntot=1,idim=1;idim<=ndim;idim++)
    ntot *= nn[idim];
  nprev=1;
  for (idim=ndim;idim>=1;idim--) {
    n=nn[idim];
    nrem=ntot/(n*nprev);
    ip1=nprev << 1;
    ip2=ip1*n;
    ip3=ip2*nrem;
    i2rev=1;
    for (i2=1;i2<=ip2;i2+=ip1) {
      if (i2 < i2rev) {
	for (i1=i2;i1<=i2+ip1-2;i1+=2) {
	  for (i3=i1;i3<=ip3;i3+=ip2) {
	    i3rev=i2rev+i3-i2;
	    SWAP(data[i3],data[i3rev]);
	    SWAP(data[i3+1],data[i3rev+1]);
	  }
	}
      }
      ibit=ip2 >> 1;
      while (ibit >= ip1 && i2rev > ibit) {
	i2rev -= ibit;
	ibit >>= 1;
      }
      i2rev += ibit;
    }
    ifp1=ip1;
    while (ifp1 < ip2) {
      ifp2=ifp1 << 1;
      theta=isign*6.28318530717959/(ifp2/ip1);
      wtemp=sin(0.5*theta);
      wpr = -2.0*wtemp*wtemp;
      wpi=sin(theta);
      wr=1.0;
      wi=0.0;
      for (i3=1;i3<=ifp1;i3+=ip1) {
	for (i1=i3;i1<=i3+ip1-2;i1+=2) {
	  for (i2=i1;i2<=ip3;i2+=ifp2) {
	    k1=i2;
	    k2=k1+ifp1;
	    tempr=(float)wr*data[k2]-(float)wi*data[k2+1];
	    tempi=(float)wr*data[k2+1]+(float)wi*data[k2];
	    data[k2]=data[k1]-tempr;
	    data[k2+1]=data[k1+1]-tempi;
	    data[k1] += tempr;
	    data[k1+1] += tempi;
	  }
	}
	wr=(wtemp=wr)*wpr-wi*wpi+wr;
	wi=wi*wpr+wtemp*wpi+wi;
      }
      ifp1=ifp2;
    }
    nprev *= n;
  }
}

// stright out of nr.  Comments would be useless :), its a magic black box function
void
fft::rlft3(float ***data, float **speq, unsigned long nn1, unsigned long nn2, unsigned long nn3, int isign)
{
  unsigned long i1,i2,i3,j1,j2,j3,nn[4],ii3;
  double theta,wi,wpi,wpr,wr,wtemp;
  float c1,c2,h1r,h1i,h2r,h2i;
  
  unsigned long toCmp = 1+&data[nn1][nn2][nn3]-&data[1][1][1];
  if (toCmp != nn1*nn2*nn3)
    nrerror("rlft3: problem with dimensions or contiguity of data array\n");
  c1=0.5;
  c2 = -0.5*isign;
  theta=isign*(6.28318530717959/nn3);
  wtemp=sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi=sin(theta);
  nn[1]=nn1;
  nn[2]=nn2;
  nn[3]=nn3 >> 1;
  if (isign == 1) {
    fourn(&data[1][1][1]-1,nn,3,isign);
    for (i1=1;i1<=nn1;i1++)
      for (i2=1,j2=0;i2<=nn2;i2++) {
	speq[i1][++j2]=data[i1][i2][1];
	speq[i1][++j2]=data[i1][i2][2];
      }
  }
  for (i1=1;i1<=nn1;i1++) {
    j1=(i1 != 1 ? nn1-i1+2 : 1);
    wr=1.0;
    wi=0.0;
    for (ii3=1,i3=1;i3<=(nn3>>2)+1;i3++,ii3+=2) {
      for (i2=1;i2<=nn2;i2++) {
	if (i3 == 1) {
	  j2=(i2 != 1 ? ((nn2-i2)<<1)+3 : 1);
	  h1r=c1*(data[i1][i2][1]+speq[j1][j2]);
	  h1i=c1*(data[i1][i2][2]-speq[j1][j2+1]);
	  h2i=c2*(data[i1][i2][1]-speq[j1][j2]);
	  h2r= -c2*(data[i1][i2][2]+speq[j1][j2+1]);
	  data[i1][i2][1]=h1r+h2r;
	  data[i1][i2][2]=h1i+h2i;
	  speq[j1][j2]=h1r-h2r;
	  speq[j1][j2+1]=h2i-h1i;
	} else {
	  j2=(i2 != 1 ? nn2-i2+2 : 1);
	  j3=nn3+3-(i3<<1);
	  h1r=c1*(data[i1][i2][ii3]+data[j1][j2][j3]);
	  h1i=c1*(data[i1][i2][ii3+1]-data[j1][j2][j3+1]);
	  h2i=c2*(data[i1][i2][ii3]-data[j1][j2][j3]);
	  h2r= -c2*(data[i1][i2][ii3+1]+data[j1][j2][j3+1]);
	  data[i1][i2][ii3]=h1r+wr*h2r-wi*h2i;
	  data[i1][i2][ii3+1]=h1i+wr*h2i+wi*h2r;
	  data[j1][j2][j3]=h1r-wr*h2r+wi*h2i;
	  data[j1][j2][j3+1]= -h1i+wr*h2i+wi*h2r;
	}
      }
      wr=(wtemp=wr)*wpr-wi*wpi+wr;
      wi=wi*wpr+wtemp*wpi+wi;
    }
  }
  if (isign == -1)
    fourn(&data[1][1][1]-1,nn,3,isign);
}

#undef SWAP





