/**************************************************************************

   File:                kdtree.h

   Author(s):           Deeptiman Jugessur

   Created:             23 Jun 1999

   Last Revision:       $Date$

   Description: 

   $Revision$

   $Log$

	
   Copyright (c) 1999 by Deeptiman Jugessur, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#ifndef KDTREE_H
#  define KDTREE_H

#include <assert.h>

// requires the code that instantiates the kdtree to have two functions, 
// keyAtFn and distFn.
template<class T, float (*keyAtFn)(const T*,const int), float (*distFn)(const T*, const T*)>
class kdtree
{
private:
  struct kdnode
  {
    T *data;
    struct kdnode *leftChild;
    struct kdnode *rightChild;
    int index; // in case you want to keep track of where things go in the tree
    int visited; // to mark if a node has been visited by getClosest function
    kdnode(T *d=NULL, kdnode *l=NULL, kdnode *r=NULL, int i=-1, int j=0) : data(d), leftChild(l), rightChild(r), index(i), visited(j) {} 
    ~kdnode() {
      if (leftChild != NULL) {
	// cout << "Deleting left child\n";
	delete leftChild;
	leftChild = NULL;
      }
      if (rightChild != NULL) {
	// cout << "Deleting right child\n";
	delete rightChild;
	rightChild = NULL;
      }
      if (data != NULL) {
	// cout << "Deleting data\n";
	delete data; 
	data = NULL;
      }
    }
  };

  struct kdnode *root;                // root of the kd-tree
  float minDist;                      // holds distance to closest point found so far in search
  struct kdnode *minNode;            // holds pointer to closest point found so far in search
  struct kdnode **minNodes;          // hold the initial N min nodes.
  float *minDists;                   // holds distances to closest points found so far in search for N nodes

public:
  int numAdd;
  int numDel;
  kdtree(){numAdd=0;numDel=0;root=NULL;}
  ~kdtree(){
    cout << "Destroying kdtree with: " << numAdd << " nodes\n";
    actualDestroy(root,0);
    cout << "kdtree dead, killed " << numDel << " nodes\n";
  }
  void insert(T* t, int index=-1) {
    struct kdnode *newNode = new kdnode(t,NULL,NULL,index);
    //    cout << "Inserting in tree:" << *(newNode->data);
    numAdd++;
    if (root != NULL)
      actualInsert(newNode,0,root);
    else
      root = newNode;
  };

  void print() {
    actualPrint(root,0);
  };    

  T* findClosest(T* t, int *where=NULL, int markVisited=0) {
    minNode = root;
    minDist=distFn(root->data,t);
    actualFindClosest(root,t,0,markVisited);
    if (where != NULL)
      *where = minNode->index;
    if (markVisited)
      minNode->visited = 1;
    //    cout << "KDT says where = " << *where << "\n";
    return minNode->data;
  };


  // *where is assumed to be a pre-allocated array of N ints
  T** findNClosest(T* t, int N, int *where=NULL) {
    assert(N <= numAdd);
    
    T** toReturn;
    
    // allocate space for minNodes and toReturn + initialize
    minNodes = new kdnode*[N];
    
    toReturn = new T*[N];
    for (int i=0; i<N; i++)
      toReturn[i] = new T;
    minDists = new float[N];
    
    for (int i=0; i<N; i++) {
      minNodes[i] = NULL;
      minDists[i] = -1;
    }

    // initialize minNodes (just get N unique ones by by doing a post order traversal)
    getNNodes(minNodes,N);

   
    // initialize minDists and indeces 
    for (int i=0; i<N; i++) {
      minDists[i] = distFn(minNodes[i]->data,t);
      if (where)
	where[i] = minNodes[i]->index;
    }

    // initialize minDist to max of dists in array
    minDist = getMax(minDists,N);

    // the culprit!  
    actualFindNClosest(root,t,0,N);  // fill minNodes with the N closest neighbors
    
    for (int i=0; i<N; i++) {
      //      *(toReturn[i]) = *(minNodes[i]->data);  // assumes data type has copy constructor operator=
      toReturn[i]->copyVector(minNodes[i]->data);
      if (where) where[i] = minNodes[i]->index;
    }
    
    delete []minDists;
    for (int i=0; i<N; i++)
      minNodes[i] = NULL;

    delete []minNodes;
    return toReturn;
  };



  // *where is assumed to be a pre-allocated array of N ints
  void findNClosest(T** toFill, T* t, int N, int *where=NULL) {
    assert(N <= numAdd);
    
    // allocate space for minNodes and toReturn + initialize
    minNodes = new kdnode*[N];
    minDists = new float[N];
    
    for (int i=0; i<N; i++) {
      minNodes[i] = NULL;
      minDists[i] = -1;
    }

    // initialize minNodes (just get N unique ones by by doing a post order traversal)
    getNNodes(minNodes,N);

   
    // initialize minDists and indeces 
    for (int i=0; i<N; i++) {
      minDists[i] = distFn(minNodes[i]->data,t);
      if (where)
	where[i] = minNodes[i]->index;
    }

    // initialize minDist to max of dists in array
    minDist = getMax(minDists,N);

    // the culprit!  
    actualFindNClosest(root,t,0,N);  // fill minNodes with the N closest neighbors
    
    for (int i=0; i<N; i++) {
      //      *(toReturn[i]) = *(minNodes[i]->data);  // assumes data type has copy constructor operator=
      toFill[i]->copyVector(minNodes[i]->data);
      if (where) 
	where[i] = minNodes[i]->index;
    }
    
    delete []minDists;
    for (int i=0; i<N; i++)
      minNodes[i] = NULL;

    delete []minNodes;
  };

  int getNumNodes() {return numAdd;}
  
  // perform a post order traversal to get the first N nodes
  void getNNodes(struct kdnode **nodes, int N) {    
    assert(N < numAdd);
    int currN = N;
    //    actualPostOrder(nodes,root,0,N);
    actualPreOrder(nodes,root,&currN);
    if (currN != 0)
      cout << "SEVERE Error: Not enough elements in tree to initialize N-closest array." << endl;
  };


  float getMax(float *arr, int N, int *index=NULL) {
    float max = arr[0];
    if (index) *index=0;
    for (int i=0; i<N; i++) {
      if (arr[i] > max) {
	max = arr[i];
	if (index != NULL)
	  *index = i;
      }
    }
    return max;
  };
  
  
private:
  int actualPostOrder(struct kdnode **nodes, struct kdnode *it, int whereIndex, int N) {
    if (it == NULL)
      return whereIndex;
    
    //    cout << "COUNT = " << count << "\n";
    whereIndex = actualPostOrder(nodes,it->leftChild,whereIndex,N);
    whereIndex = actualPostOrder(nodes,it->rightChild,whereIndex,N);

    if (whereIndex < N) {
      nodes[whereIndex++] = it;      
    }

    return whereIndex;
  }
  
  void actualPreOrder(struct kdnode **nodes,struct kdnode *root,int *n){
    if (root != NULL && *n>0){
      *n = *n - 1;
      nodes[*n] = root;
      actualPreOrder(nodes,root->leftChild,n);
      actualPreOrder(nodes,root->rightChild,n);
    }
  }


  void actualDestroy(struct kdnode *it, int level) {
    // postorder deletion of each node
    if (it == NULL)
      return;

    actualDestroy(it->leftChild, level+1);
    actualDestroy(it->rightChild, level+1);

    if (it->leftChild != NULL) {
      it->leftChild=NULL;
    }
    if (it->rightChild != NULL) {
      it->rightChild=NULL;
    }

    delete it;
    it=NULL;

    numDel++;

  }
    

  void  actualFindClosest(struct kdnode* r, T* t, int level, int checkVisited) {
    if (r==NULL)
      return;
    float dist=distFn(r->data,t);
    if (!checkVisited) {
      if (dist < minDist) {
	minDist = dist;
	minNode = r;
      }
    }
    else if (dist < minDist && !(r->visited)) {
      minDist = dist;
      minNode = r;
    }  

    dist = keyDist(r,t,level);
    if ( dist > minDist ){   // search only sub-branch closer to point 't'
      if ( keyAtFn(t,level) < keyAtFn(r->data,level) )
	actualFindClosest(r->leftChild,t,level+1,checkVisited);
      else
	actualFindClosest(r->rightChild,t,level+1,checkVisited);
    }
    else{   // search both sub-branches
      actualFindClosest(r->leftChild,t,level+1,checkVisited);
      if ( dist < minDist )
	actualFindClosest(r->rightChild,t,level+1,checkVisited);
    }
  }


  void  actualFindNClosest(struct kdnode* r, T* t, int level, int N) {
    if (r==NULL)
      return;
    float dist=distFn(r->data,t);
    if (dist < minDist) {
      // get the index of the minNode that gave the min dist (ie the max one)
      int maxIndex=-1;
      getMax(minDists,N,&maxIndex);
      // insert r into minNodes
      minNodes[maxIndex] = r;
      // update minDists array
      minDists[maxIndex] = dist;
      // update minDist
      minDist = getMax(minDists,N);      
    }

    dist = keyDist(r,t,level);
    if ( dist > minDist ){   // search only sub-branch closer to point 't'
      if ( keyAtFn(t,level) < keyAtFn(r->data,level) )
	actualFindNClosest(r->leftChild,t,level+1,N);
      else
	actualFindNClosest(r->rightChild,t,level+1,N);
    }
    else{   // search both sub-branches
      actualFindNClosest(r->leftChild,t,level+1,N);
      if ( dist < minDist )
	actualFindNClosest(r->rightChild,t,level+1,N);
    }
  }
  
  float keyDist(struct kdnode* a,T *b,int level){
    return abs(keyAtFn(a->data,level)-keyAtFn(b,level));
  }

  void actualPrint(struct kdnode *it,int level){
    if (it == NULL) 
      return;
    cout << "L";
    actualPrint(it->leftChild,level + 1);
    cout << "U";
    cout << " Node (level " << level << ") : " << *(it->data);
    cout << "R";
    actualPrint(it->rightChild,level + 1);
    cout << "U";
  }
  
  void actualInsert(struct kdnode *newN,int level,struct kdnode *r){
    if (keyAtFn(newN->data,level) < keyAtFn(r->data,level)){  // insert left
      if (r->leftChild == NULL) {
	r->leftChild = newN;
      }
      else
	actualInsert(newN,level+1,r->leftChild);
    }
    else { // insert right ie >=
      if (r->rightChild == NULL) {
	r->rightChild = newN;
      }
      else
	actualInsert(newN,level+1,r->rightChild);
    }
  };
};



#endif  /* KDTREE_H */






