// Author: Ariel Tankus.
// Created: Sept. 1996.
// Last modified: 26.3.1999.

#include <math.h>
#include "mex.h"
#include "general.h"
#include "matrix_ifc.h"

/* Input Arguments */
#define	IM_IN		prhs[0]

/* Output Arguments */
#define	LIGHTCOR_OUT	0

//
// Fits a linear function (under least squares approximation) to the given
// Fimage. The returned image contains the values of the linear function.
//
// Motivation: Subtracting the linear regression from the original function
// might be helpful in eliminating illumination conditions. More specifically,
// the linear regression reduces the effect of shadow introduced by a single
// lateral source of light directed at the subject.
//
void correctLighting(const mymatrix<double>& im, mymatrix<double>& result)
{
    int width  = im.columns();
    int height = im.rows();
    int size   = im.size();

    double sumXZ = 0;
    double sumYZ = 0;
    double sumXY = 0;
    double sumXX = 0;
    double sumYY = 0;
    double sumZ  = 0;
    int x, y;
    for (x = 0; x < width; x++)
	for (y = 0; y < height; y++) {
	    sumZ += im(y, x);
	    sumXZ += x * im(y, x);
	    sumYZ += y * im(y, x);
	    sumXY += x * y;
	    sumXX += x * x;
	    sumYY += y * y;
	}

    double sumX = height*width*(width-1) / 2;
    double sumY = width*height*(height-1) / 2;

    double BXY = sumXY - sumX*sumY/size;
    double BXZ = sumXZ - sumX*sumZ/size;
    double BYZ = sumYZ - sumY*sumZ/size;
    double BXX = sumXX - sumX*sumX/size;
    double BYY = sumYY - sumY*sumY/size;
    double gama = (BYZ*BXX - BXY*BXZ) / (BYY*BXX - BXY*BXY);
    double beta = (BXZ*BYY - BXY*BYZ) / (BYY*BXX - BXY*BXY);
    double alpha = (sumZ - sumX*beta - sumY*gama) / size;

    double *linearArray = new double[size];
    for (x = 0; x < width; x++)
	for (y = 0; y < height; y++)
	    result(y, x) = alpha + beta * x + gama * y;
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    /* Check for proper number of arguments */
    if (nrhs != 1)
	mexErrMsgTxt("RADIAL requires one input argument.");
    else if (nlhs > 1)
	mexErrMsgTxt("RADIAL requires one output arguments at most.");
    
    /* Check the input dimensions. */
    if (!isNormal(IM_IN))
	mexErrMsgTxt("CORLIGHT requires that the input matrix is a real, numeric, full, double matrix.");


    /* Assign pointer to the parameter */
    mymatrix<double> im = MatrixTomatrix(IM_IN);

    mymatrix<double> *out = createOutput(nlhs, plhs, LIGHTCOR_OUT, im.rows(),
					 im.columns());

    /* Do the actual computations in a subroutine */
    correctLighting(im, *out);

    return;
}
