// Author: Ariel Tankus.
// Created: Sept. 1996.
// Last modified: 26.3.1999.

#include <math.h>
#include "mex.h"
#include "general.h"
#include "matrix_ifc.h"

/* Input Arguments */
#define	RX_IN		prhs[0]
#define	RY_IN		prhs[1]
#define	KIND_IN		prhs[2]
#define	DX_IN		prhs[3]
#define	DY_IN		prhs[4]
#define	MAG_IN		prhs[5]
#define	MAGLOG_IN	prhs[6]
#define	ARG_IN  	prhs[7]

/* Output Arguments */
#define	SYMRAD_OUT	0
#define	SYMMAG_OUT	1
#define	SYMARG_OUT	2

enum symmetry_kind {Full,Dark,Bright};
const double EPSILON = 1E-13;

//
// Computes the radial symmetry map of an image, as well as magnitude and arg
// of isotropic symmetry map. Capable of computing full/dark/bright symmetries.
//
// Input:
// ------
// rx, ry     - radii for the neighborhood where symmetry is considered.
// kind       - the kind of symmetry searched: full, dark or bright.
// dxIm, dyIm - derivatives of the image in x, y directions respectively.
// gradIm, logGradIm - magnitude of gradient (and it's log) of the images.
// argIm      - arg of the gradient to the image.
//
// Output:
// -------
// symMagnit - magnitude of the symmetry map.
// symOrient - orientation of the symmetry map.
//
//    if either `symMagnit' or `symOrient' equals NULL, that parameter is left
// unchanged.
//
// Return value:
// -------------
// radial symmetry map.
//
void fastRadialSymmetry(int rx, int ry, symmetry_kind kind,
			const mymatrix<double>& dxIm,
			const mymatrix<double>& dyIm,
			const mymatrix<double>& gradIm,
			const mymatrix<double>& logGradIm,
			const mymatrix<double>& argIm,
			mymatrix<double> *symRadial,
			mymatrix<double> *symMag,
			mymatrix<double> *symOrient)
{
    int width  = gradIm.columns();
    int height = gradIm.rows();


    const double *grad = (const double *)gradIm;
    const double *arg = (const double *)argIm;
    const double *logGrad = (const double *)logGradIm;

    int mx = 2*rx+1;
    int my = 2*ry+1;
    double sqr_ry = ry*ry;
    double *EXP = new double[(mx+1)*(my+1)];
    double factor = 0.25*(-3)/(mx*mx+my*my);

    // create exp table.
    // (i,j) = x-distance and y-distance
    // between points in current rectangle
    register int i,j;
    for (i = 0; i <= mx; i++)
	for (j = 0; j <= my; j++)
	    EXP[i + (mx+1)*j] = exp(factor*(i*i+j*j));

    mymatrix<double>* cosIm = new mymatrix<double>(height, width);
    mymatrix<double>* sinIm = new mymatrix<double>(height, width);
    dxIm.elemDiv(gradIm, *cosIm);
    dyIm.elemDiv(gradIm, *sinIm);

    const double *cosTable = (const double *)*cosIm;
    const double *sinTable = (const double *)*sinIm;

    // (centerx, centery) = coordinate of center of current
    //			    rectangular environment
    // p, q = indices in the image array (treated as 1D) of
    //        current points in current rectangle.
    // o = index in the array of center of current rectangle.
    int centerx, centery;
    int x, y;		// (x, y) = current point in current rectangle
    int p, q, o;
    int deltaX, deltaY;
    double val;		// current symmetry value (for given p, q)

    double *symMagCos = clear(width*height);
    double *symMagSin = clear(width*height);
    double *maxSym    = clear(width*height);	// maximal sym mag value

    // initialize output parameters
    symRadial->setAllElem(0);
    int isOrient = (symOrient != NULL);
    if (!isOrient)
	symOrient = new mymatrix<double>(height, width);
    symOrient->setAllElem(0);
    int isMag = (symMag != NULL);
    if (!isMag)
	symMag = new mymatrix<double>(height, width);
    symMag->setAllElem(0);

    int envY;		// maximal y in env.
    int elliptic_x;	// maximal x in env. to create elliptic env.

    for (centery = 0; centery < height; centery++)
	for (centerx = 0; centerx < width; centerx++) {

	    o = height*centerx + centery;

	    // ensure y inside image boundaries; usually: envY == ry .
	    envY = min(min(ry, centery), height-centery-1);

	    for (y = -envY; y <= envY; y++) {

		// elliptic environment
		elliptic_x = int(rx*sqrt(1-y*y/sqr_ry));

		// ensure env within boundaries
		elliptic_x = min(min(elliptic_x, centerx), width-centerx-1);

		// iterate elliptic env. disregard center of env. (x==y==0)
		for (x = 0; x <= elliptic_x; x++) {
		    if (x == 0 || y == 0)
			continue;

			    q = o + (height*x + y);
			    p = o - (height*x + y);
			    
			    if (grad[p] == 0 || grad[q] == 0)
				continue;
			    
			    // (deltaX, deltaY) = q-p, where p<q
			    deltaX = 2*x+1;
			    deltaY = 2*y+1;
			    
			    if (kind == Dark &&
				(deltaX*cosTable[p]+deltaY*sinTable[p] > 0 ||
				 deltaX*cosTable[q]+deltaY*sinTable[q] < 0))
				continue;

			    if (kind == Bright &&
				(deltaX*cosTable[p]+deltaY*sinTable[p] < 0 ||
				 deltaX*cosTable[q]+deltaY*sinTable[q] > 0))
				continue;

			    val = deltaX * (cosTable[p] - cosTable[q]) +
				deltaY * (sinTable[p] - sinTable[q]);
			    val *= val;
			    val /= (deltaX*deltaX+deltaY*deltaY);
			    val *= EXP[deltaX + (mx+1)*abs(deltaY)] *
                                      logGrad[p] * logGrad[q];

			    if (val < 0)
				cerr << "val < 0\n";

			    (*symMag)(centery, centerx) += val;
			    // \Sum_{(p,q) \in \Gama(o)} 
			    //     { C(p,q) \Cos(\theta(p)+\theta(q)) }
			    symMagCos[o] += val * (cosTable[p] * cosTable[q] -
						   sinTable[p] * sinTable[q]);
			    // \Sum_{(p,q) \in \Gama(o)} 
			    //     { C(p,q) \Sin(\theta(p)+\theta(q)) }
			    symMagSin[o] += val * (sinTable[p] * cosTable[q] +
						   sinTable[q] * cosTable[p]);
			    if (val > maxSym[o]) {
				maxSym[o] = val;
				(*symOrient)(centery, centerx) = (arg[p] + arg[q]) / 2;
			    }
		}
	    }

	    (*symRadial)(centery, centerx) = ((*symMag)(centery, centerx) -
	             cos(2*(*symOrient)(centery, centerx)) * symMagCos[o] -
	             sin(2*(*symOrient)(centery, centerx)) * symMagSin[o]) / 2;

	    if ((*symRadial)(centery, centerx) < 0)
		if ((*symRadial)(centery, centerx) > -EPSILON)
		    (*symRadial)(centery, centerx) = 0;
	}

    if (!isOrient)
	delete symOrient;
    if (!isMag)
	delete symMag;

    delete [] maxSym;
    delete [] symMagCos;
    delete [] symMagSin;
    delete sinIm;
    delete cosIm;
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    /* dimensions. */
    unsigned int dx_m, dx_n, dy_m, dy_n, mag_m, mag_n, maglog_m, maglog_n;
    unsigned int arg_m, arg_n;

    /* Check for proper number of arguments */
    if (nrhs != 8)
	mexErrMsgTxt("RADIAL requires eight input arguments.");
    else if (nlhs > 3)
	mexErrMsgTxt("RADIAL requires three output arguments at most.");
    
    /* Check the input dimensions. */
    dx_m = mxGetM(DX_IN);
    dx_n = mxGetN(DX_IN);
    dy_m = mxGetM(DY_IN);
    dy_n = mxGetN(DY_IN);
    mag_m = mxGetM(MAG_IN);
    mag_n = mxGetN(MAG_IN);
    maglog_m = mxGetM(MAGLOG_IN);
    maglog_n = mxGetN(MAGLOG_IN);
    arg_m = mxGetM(ARG_IN);
    arg_n = mxGetN(ARG_IN);
    if (!isNormal(DX_IN) || !isNormal(DY_IN) || !isNormal(MAG_IN) ||
	!isNormal(MAGLOG_IN) || !isNormal(ARG_IN) ||
	(dx_m != dy_m) || (dx_m != mag_m) || (dx_m != maglog_m) || 
	(dx_m != arg_m) || 
	(dx_n != dy_n) || (dx_n != mag_n) || (dx_n != maglog_n) || 
	(dx_n != arg_n))
	mexErrMsgTxt("RADIAL requires that DX, DY, MAG, MAGLOG and ARG be matrices of the same size.");


    /* Assign pointers to the various parameters */
    int rx = MatrixToInt(RX_IN);
    int ry = MatrixToInt(RY_IN);
    symmetry_kind kind = (symmetry_kind)MatrixToInt(KIND_IN);
    mymatrix<double> dx     = MatrixTomatrix(DX_IN);
    mymatrix<double> dy     = MatrixTomatrix(DY_IN);
    mymatrix<double> mag    = MatrixTomatrix(MAG_IN);
    mymatrix<double> maglog = MatrixTomatrix(MAGLOG_IN);
    mymatrix<double> arg  = MatrixTomatrix(ARG_IN);

    mymatrix<double> *symRadial = createOutput(nlhs, plhs, SYMRAD_OUT,
					       dx.rows(), dx.columns());
    mymatrix<double> *symMag    = createOutput(nlhs, plhs, SYMMAG_OUT,
					       dx.rows(), dx.columns());
    mymatrix<double> *symOrient = createOutput(nlhs, plhs, SYMARG_OUT,
					       dx.rows(), dx.columns());

    /* Do the actual computations in a subroutine */
    fastRadialSymmetry(rx, ry, kind, dx, dy, mag, maglog, arg, symRadial,
		       symMag, symOrient);

    return;
}
