#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef __STDC__
#define error_macro(a) \
{ \
  fprintf (stderr,"An error occured in %s line %d: %s\n", \
	  __FILE__, \
	  __LINE__, \
	  a); \
  exit(1); \
}
#else
#error Standard C compiler required.
#endif __STDC__

#define PRECISION float /* a handy way of changing the precision */
#define MATRIX PRECISION**
#define VECTOR PRECISION*
#define FILE_NAME_SIZE 128

/* prototypes */
MATRIX read_matrix (char *filename);
VECTOR read_vector (char *filename);
void free_matrix (MATRIX A);
void free_vector (VECTOR x);
void print_matrix (MATRIX A);
void print_vector (VECTOR x);
void solve_lower_triangular (MATRIX L, VECTOR b);
void solve_lower_triangular_as_upper (MATRIX L, VECTOR b);
PRECISION vec_norm_2 (VECTOR x);
PRECISION relative_error (VECTOR x);
PRECISION frobenius_norm (MATRIX A);
VECTOR vector_copy (VECTOR x);
VECTOR matrix_times_vector (MATRIX A, VECTOR x);
MATRIX compute_l_lt (MATRIX L);
PRECISION matrix_residual (MATRIX A, MATRIX L);
MATRIX cholesky_factorize (MATRIX A);

/* definitions */
MATRIX read_matrix (char *filename){
  FILE *mat_file;
  int n;
  int i,j;
  MATRIX A;

  mat_file=fopen(filename,"r"); /* open the file containing the matrix */
  if (!mat_file)
    error_macro("file does not exist");
  /* now determine the size of the matrix */
  /* only need one dimension since we know matrix is SPD */
  fread(&n,sizeof(int),1,mat_file);
  /* now allocate enough space */
  A=(PRECISION**)malloc((n+1)*sizeof(PRECISION*)); /* use math bounds instead of C */
  if (!A)
    error_macro("malloc failed");
  for (i=1; i<=n; i++){
    A[i]=(PRECISION*)malloc((n+1)*sizeof(PRECISION));
    if (!A[i])
      error_macro("malloc failed");
  }
  /* now we have enough space to store the matrix, read it in */
  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      fread(&(A[i][j]),sizeof(PRECISION),1,mat_file);
  fclose(mat_file);
  /* store size in first pointer */
  A[0]=(PRECISION*)n;
  return A;
}

VECTOR read_vector (char *filename){
  FILE *vec_file;
  int n;
  int i;
  VECTOR x;

  vec_file=fopen(filename,"r"); /* open the file containing the vector */
  if (!vec_file)
    error_macro("file does not exist");
  /* now determine the size of the vector */
  fread(&n,sizeof(int),1,vec_file);
  /* now allocate enough space */
  x=(PRECISION*)malloc((n+1)*sizeof(PRECISION)); /* use math bounds instead of C */
  if (!x)
    error_macro("malloc failed");
  /* now we have enough space to store the vector, read it in */
  for (i=1; i<=n; i++)
    fread(&(x[i]),sizeof(PRECISION),1,vec_file);
  fclose(vec_file);
  /* store size in first pointer */
  x[0]=(PRECISION)n;
  return x;
}

void free_matrix (MATRIX A){
  int n=(int)A[0];
  int i;
  
  for (i=1; i<=n; i++)
    free(A[i]);
}

void free_vector (VECTOR x){
  free(x);  
}

void print_matrix (MATRIX A){
  int n=(int)A[0];
  int i,j;
  
  for (i=1; i<=n; i++){
    for (j=1; j<=n; j++){
      printf ("%f\t",A[i][j]);
    }
    printf("\n");
  }
}

void print_vector (VECTOR x){
  int n=(int)x[0];
  int i;

  for (i=1; i<=n; i++)
    printf ("%f\n",x[i]);
}

void solve_lower_triangular (MATRIX L, VECTOR b){
  int i,j,n;
  /* get the dimension of A */
  n=(int)L[0];
  /* now proceed and solve for y */
  for(i=1; i<=n; i++){
    for (j=1; j<i; j++)
      b[i]=b[i]-L[i][j]*b[j];
    if (L[i][i]==0){
      error_macro("divide by zero");
    }
    else
      b[i]=b[i]/L[i][i];
  }
}

void solve_lower_triangular_as_upper (MATRIX L, VECTOR b){
  int i,j,n;
  /* get the dimension of A */
  n=(int)L[0];
  /* now proceed and solve for y */
  for(j=n; j>=1; j--){
    for (i=j+1; i<=n; i++)
      b[j]=b[j]-L[i][j]*b[i];
    if (L[j][j]==0){
      error_macro("divide by zero");
    }
    else
      b[j]=b[j]/L[j][j];
  }
}

MATRIX cholesky_factorize (MATRIX A){
  int i,j,k,n;
  MATRIX L;
  n=(int)A[0]; /* get the size of A */
  /* allocate matrix L */
  L=(PRECISION**)malloc((n+1)*sizeof(PRECISION*));
  if (!L)
    error_macro("malloc failed");
  for (i=1; i<=n; i++){
    L[i]=(PRECISION*)malloc((n+1)*sizeof(PRECISION));
    if (!L[i])
      error_macro("malloc failed");
  }
  /* set the size */
  L[0]=(PRECISION*)n;
  /* now do the factorization */
  for (j=1; j<=n; j++)
    for (i=j; i<=n; i++){
      L[i][j]=A[i][j];
      for (k=1; k<j; k++)
	L[i][j]=L[i][j]-L[i][k]*L[j][k];
      if (i==j){
	if (L[i][j]<0){
	  error_macro("square root of negative number");
	}
	else
	  L[i][j]=sqrt(L[i][j]);
      }
      else{
	if (L[j][j]==0){
	  error_macro("divide by zero");
	}
	else
	  L[i][j]/=L[j][j];
      }
    }
  return L; /* L does not have zeros in upper part */
}

PRECISION vec_norm_2 (VECTOR x){
  PRECISION norm=0;
  int i;
  int n=(int)x[0];
  
  for (i=1; i<=n; i++)
    norm+=x[i]*x[i];

  return(sqrt(norm));
}

PRECISION relative_error (VECTOR x){
  int i;
  int n=(int)x[0];
  PRECISION norm_xc;
  PRECISION relative_err;

  norm_xc=vec_norm_2(x);
  for (i=1; i<=n; i++)
    x[i]-=1; /* subtract e */
  relative_err=vec_norm_2(x)/norm_xc; /* ||xc-e|| / ||xc|| */
  /* add e to x to restore (faster than a copy */
  for (i=1; i<=n; i++)
    x[i]+=1;
  return relative_err;
}

PRECISION frobenius_norm (MATRIX A){
  int i,j,n=(int)A[0];
  PRECISION norm=0;

  for (i=1;i<=n;i++)
    for (j=1;j<=n;j++)
      norm+=A[i][j]*A[i][j];
  if (norm<0){
    error_macro("square root of negative");
  }
  else
    norm=sqrt(norm);
  return norm;
}

VECTOR vector_copy (VECTOR x){
  PRECISION *y;
  int i,n=(int)x[0];
  y=(PRECISION*)malloc((n+1)*sizeof(PRECISION));
  if (!y)
    error_macro("malloc failed");
  for (i=1; i<=n; i++)
    y[i]=x[i];
  y[0]=n;
  return y;
}

VECTOR matrix_times_vector (MATRIX A, VECTOR x){
  PRECISION *y;
  int i,j,n=(int)A[0];
  y=(PRECISION*)malloc((n+1)*sizeof(PRECISION));
  if (!y)
    error_macro("malloc failed");
  /* now do the multiplication */
  for (i=1; i<=n; i++){
    y[i]=0;
    for (j=1; j<=n; j++)
      y[i]+=A[i][j]*x[j];
  }
  y[0]=(int)n;
  return y;
}

PRECISION relative_residual (MATRIX A, VECTOR xc, VECTOR b){
  VECTOR Axc;
  int i,n=(int)A[0];
  PRECISION residual;
  PRECISION normalizer;

  Axc=matrix_times_vector(A,xc);
  for (i=1; i<=n; i++)
    b[i]-=Axc[i]; /* compute b-Axc and store in b */
  residual=vec_norm_2(b);
  normalizer=frobenius_norm(A)*vec_norm_2(xc);
  if (normalizer==0){
    error_macro("divide by zero");
  }
  residual/=normalizer;
  free_vector(Axc);
  return residual;
}

MATRIX compute_l_lt (MATRIX L){
  int i,j,k,n=(int)L[0];
  MATRIX Ac;

  /* allocate the rows */
  Ac=(PRECISION**)malloc((n+1)*sizeof(PRECISION*));
  if (!Ac)
    error_macro("malloc failed");
  /* now the columns */
  for (i=1; i<=n; i++){
    Ac[i]=(PRECISION*)malloc((n+1)*sizeof(PRECISION));
    if (!Ac[i])
      error_macro("malloc failed");
  }
  Ac[0]=(PRECISION*)n; /* save the size */
  /* now compute L*Lt */
  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++){
      Ac[i][j]=0;
      for (k=1; k<=((i<j)?i:j); k++)
	Ac[i][j]+=L[i][k]*L[j][k];
    }
  return Ac;
}

PRECISION matrix_residual (MATRIX A, MATRIX L){
  MATRIX Ac;
  PRECISION frob_norm_A;
  PRECISION residual;
  int i,j,n=(int)A[0];
  
  Ac=compute_l_lt(L);
  frob_norm_A=frobenius_norm(A);
  /* calculate Ac - A into Ac*/
  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      Ac[i][j]-=A[i][j];
  residual=frobenius_norm(Ac)/frob_norm_A;
  free_matrix(Ac);
  return(residual);
}

int main (int argc, char *argv[]){
  char filename[FILE_NAME_SIZE];
  MATRIX A;
  MATRIX L;
  VECTOR b;
  VECTOR xc;
  
  if (argc!=2){
    printf("Usage: q1 prefix_name\n");
    exit(1);
  }
  sprintf(filename,"%s.mat",argv[1]);
  A=read_matrix(filename);
  sprintf(filename,"%s.vec",argv[1]);
  b=read_vector(filename);

  printf("Now performing Cholesky factorization on Ax=b with A:\n\n");
  print_matrix(A);
  printf("\nand b:\n");
  print_vector(b);
  L=cholesky_factorize(A);
  solve_lower_triangular(L,b);
  solve_lower_triangular_as_upper(L,b);
  printf ("\nSolution vector xc:\n");
  /* now compute the errors */
  /* note that the storage methods are not fancy here, but need not be fast */
  xc=vector_copy(b); /* get the computed solution {could just use b, but this is cleaner} */
  print_vector(xc);
  free_vector(b);
  b=read_vector(filename); /* get back the original b which was overwritten since we need it */
  printf ("\nRelative error: %8.3e\n",relative_error(xc));
  printf ("Relative residual: %8.3e\n",relative_residual(A,xc,b)); /* b has been changed */ 
  printf ("Relative matrix residual: %8.3e\n",matrix_residual(A,L));
  free_matrix(A);
  free_matrix(L);
  free_vector(b);
  free_vector(xc);

  return EXIT_SUCCESS;
}











